"""The base for :rfc:`5545` components."""

from __future__ import annotations

import json
from datetime import date, datetime, time, timedelta, timezone
from typing import TYPE_CHECKING, ClassVar

from icalendar.attr import (
    CONCEPTS_TYPE_SETTER,
    LINKS_TYPE_SETTER,
    RELATED_TO_TYPE_SETTER,
    comments_property,
    concepts_property,
    links_property,
    refids_property,
    related_to_property,
    single_utc_property,
    uid_property,
)
from icalendar.cal.component_factory import ComponentFactory
from icalendar.caselessdict import CaselessDict
from icalendar.error import InvalidCalendar, JCalParsingError
from icalendar.parser import Contentline, Contentlines, Parameters, q_join, q_split
from icalendar.parser_tools import DEFAULT_ENCODING
from icalendar.prop import VPROPERTY, TypesFactory, vDDDLists, vText
from icalendar.timezone import tzp
from icalendar.tools import is_date

if TYPE_CHECKING:
    from icalendar.compatibility import Self

_marker = []


class Component(CaselessDict):
    """Base class for calendar components.

    Component is the base object for calendar, Event and the other
    components defined in :rfc:`5545`. Normally you will not use this class
    directly, but rather one of the subclasses.

    Attributes:
        name: The name of the component. Example: ``VCALENDAR``.
        required: These properties are required.
        singletons: These properties must only appear once.
        multiple: These properties may occur more than once.
        exclusive: These properties are mutually exclusive.
        inclusive: If the first in a tuple occurs, the second one must also occur.
        ignore_exceptions: If True, and we cannot parse this
            component, we will silently ignore it, rather than let the
            exception propagate upwards.
        types_factory: Factory for property types
    """

    name = None  # should be defined in each component
    required = ()  # These properties are required
    singletons = ()  # These properties must only appear once
    multiple = ()  # may occur more than once
    exclusive = ()  # These properties are mutually exclusive
    inclusive: (
        tuple[str] | tuple[tuple[str, str]]
    ) = ()  # if any occurs the other(s) MUST occur
    # ('duration', 'repeat')
    ignore_exceptions = False  # if True, and we cannot parse this
    # component, we will silently ignore
    # it, rather than let the exception
    # propagate upwards
    # not_compliant = ['']  # List of non-compliant properties.

    types_factory = TypesFactory.instance()
    _components_factory: ClassVar[ComponentFactory | None] = None

    @classmethod
    def get_component_class(cls, name: str) -> type[Component]:
        """Return a component with this name.

        Arguments:
            name: Name of the component, i.e. ``VCALENDAR``
        """
        if cls._components_factory is None:
            cls._components_factory = ComponentFactory()
        return cls._components_factory.get_component_class(name)

    @staticmethod
    def _infer_value_type(
        value: date | datetime | timedelta | time | tuple | list,
    ) -> str | None:
        """Infer the ``VALUE`` parameter from a Python type.

        Args:
            value: Python native type, one of :py:class:`date`, :py:mod:`datetime`,
                :py:class:`timedelta`, :py:mod:`time`, :py:class:`tuple`,
                or :py:class:`list`.

        Returns:
            str or None: The ``VALUE`` parameter string, for example, "DATE",
                "TIME", or other string, or ``None``
                if no specific ``VALUE`` is needed.
        """
        if isinstance(value, list):
            if not value:
                return None
            # Check if ALL items are date (but not datetime)
            if all(is_date(item) for item in value):
                return "DATE"
            # Check if ALL items are time
            if all(isinstance(item, time) for item in value):
                return "TIME"
            # Mixed types or other types - don't infer
            return None
        if is_date(value):
            return "DATE"
        if isinstance(value, time):
            return "TIME"
        # Don't infer PERIOD - it's too risky and vPeriod already handles it
        return None

    def __init__(self, *args, **kwargs):
        """Set keys to upper for initial dict."""
        super().__init__(*args, **kwargs)
        # set parameters here for properties that use non-default values
        self.subcomponents: list[Component] = []  # Components can be nested.
        self.errors = []  # If we ignored exception(s) while
        # parsing a property, contains error strings

    def __bool__(self):
        """Returns True, CaselessDict would return False if it had no items."""
        return True

    def is_empty(self):
        """Returns True if Component has no items or subcomponents, else False."""
        return bool(not list(self.values()) + self.subcomponents)

    #############################
    # handling of property values

    @classmethod
    def _encode(cls, name, value, parameters=None, encode=1):
        """Encode values to icalendar property values.

        :param name: Name of the property.
        :type name: string

        :param value: Value of the property. Either of a basic Python type of
                      any of the icalendar's own property types.
        :type value: Python native type or icalendar property type.

        :param parameters: Property parameter dictionary for the value. Only
                           available, if encode is set to True.
        :type parameters: Dictionary

        :param encode: True, if the value should be encoded to one of
                       icalendar's own property types (Fallback is "vText")
                       or False, if not.
        :type encode: Boolean

        :returns: icalendar property value
        """
        if not encode:
            return value
        if isinstance(value, cls.types_factory.all_types):
            # Don't encode already encoded values.
            obj = value
        else:
            # Extract VALUE parameter if present, or infer it from the Python type
            value_param = None
            if parameters and "VALUE" in parameters:
                value_param = parameters["VALUE"]
            elif not isinstance(value, cls.types_factory.all_types):
                inferred = cls._infer_value_type(value)
                if inferred:
                    value_param = inferred
                    # Auto-set the VALUE parameter
                    if parameters is None:
                        parameters = {}
                    if "VALUE" not in parameters:
                        parameters["VALUE"] = inferred

            klass = cls.types_factory.for_property(name, value_param)
            obj = klass(value)
        if parameters:
            if not hasattr(obj, "params"):
                obj.params = Parameters()
            for key, item in parameters.items():
                if item is None:
                    if key in obj.params:
                        del obj.params[key]
                else:
                    obj.params[key] = item
        return obj

    def add(
        self,
        name: str,
        value,
        parameters: dict[str, str] | Parameters = None,
        encode: bool = True,  # noqa: FBT001
    ):
        """Add a property.

        :param name: Name of the property.
        :type name: string

        :param value: Value of the property. Either of a basic Python type of
                      any of the icalendar's own property types.
        :type value: Python native type or icalendar property type.

        :param parameters: Property parameter dictionary for the value. Only
                           available, if encode is set to True.
        :type parameters: Dictionary

        :param encode: True, if the value should be encoded to one of
                       icalendar's own property types (Fallback is "vText")
                       or False, if not.
        :type encode: Boolean

        :returns: None
        """
        if isinstance(value, datetime) and name.lower() in (
            "dtstamp",
            "created",
            "last-modified",
        ):
            # RFC expects UTC for those... force value conversion.
            value = tzp.localize_utc(value)

        # encode value
        if (
            encode
            and isinstance(value, list)
            and name.lower() not in ["rdate", "exdate", "categories"]
        ):
            # Individually convert each value to an ical type except rdate and
            # exdate, where lists of dates might be passed to vDDDLists.
            value = [self._encode(name, v, parameters, encode) for v in value]
        else:
            value = self._encode(name, value, parameters, encode)

        # set value
        if name in self:
            # If property already exists, append it.
            oldval = self[name]
            if isinstance(oldval, list):
                if isinstance(value, list):
                    value = oldval + value
                else:
                    oldval.append(value)
                    value = oldval
            else:
                value = [oldval, value]
        self[name] = value

    def _decode(self, name, value):
        """Internal for decoding property values."""

        # TODO: Currently the decoded method calls the icalendar.prop instances
        # from_ical. We probably want to decode properties into Python native
        # types here. But when parsing from an ical string with from_ical, we
        # want to encode the string into a real icalendar.prop property.
        if isinstance(value, vDDDLists):
            # TODO: Workaround unfinished decoding
            return value
        decoded = self.types_factory.from_ical(name, value)
        # TODO: remove when proper decoded is implemented in every prop.* class
        # Workaround to decode vText properly
        if isinstance(decoded, vText):
            decoded = decoded.encode(DEFAULT_ENCODING)
        return decoded

    def decoded(self, name, default=_marker):
        """Returns decoded value of property."""
        # XXX: fail. what's this function supposed to do in the end?
        # -rnix

        if name in self:
            value = self[name]
            if isinstance(value, list):
                return [self._decode(name, v) for v in value]
            return self._decode(name, value)
        if default is _marker:
            raise KeyError(name)
        return default

    ########################################################################
    # Inline values. A few properties have multiple values inlined in in one
    # property line. These methods are used for splitting and joining these.

    def get_inline(self, name, decode=1):
        """Returns a list of values (split on comma)."""
        vals = [v.strip('" ') for v in q_split(self[name])]
        if decode:
            return [self._decode(name, val) for val in vals]
        return vals

    def set_inline(self, name, values, encode=1):
        """Converts a list of values into comma separated string and sets value
        to that.
        """
        if encode:
            values = [self._encode(name, value, encode=1) for value in values]
        self[name] = self.types_factory["inline"](q_join(values))

    #########################
    # Handling of components

    def add_component(self, component: Component):
        """Add a subcomponent to this component."""
        self.subcomponents.append(component)

    def _walk(self, name, select):
        """Walk to given component."""
        result = []
        if (name is None or self.name == name) and select(self):
            result.append(self)
        for subcomponent in self.subcomponents:
            result += subcomponent._walk(name, select)  # noqa: SLF001
        return result

    def walk(self, name=None, select=lambda _: True) -> list[Component]:
        """Recursively traverses component and subcomponents. Returns sequence
        of same. If name is passed, only components with name will be returned.

        :param name: The name of the component or None such as ``VEVENT``.
        :param select: A function that takes the component as first argument
          and returns True/False.
        :returns: A list of components that match.
        :rtype: list[Component]
        """
        if name is not None:
            name = name.upper()
        return self._walk(name, select)

    #####################
    # Generation

    def property_items(
        self,
        recursive=True,
        sorted: bool = True,  # noqa: A002, FBT001
    ) -> list[tuple[str, object]]:
        """Returns properties in this component and subcomponents as:
        [(name, value), ...]
        """
        v_text = self.types_factory["text"]
        properties = [("BEGIN", v_text(self.name).to_ical())]
        property_names = self.sorted_keys() if sorted else self.keys()

        for name in property_names:
            values = self[name]
            if isinstance(values, list):
                # normally one property is one line
                for value in values:
                    properties.append((name, value))
            else:
                properties.append((name, values))
        if recursive:
            # recursion is fun!
            for subcomponent in self.subcomponents:
                properties += subcomponent.property_items(sorted=sorted)
        properties.append(("END", v_text(self.name).to_ical()))
        return properties

    @classmethod
    def from_ical(cls, st, multiple: bool = False) -> Self | list[Self]:  # noqa: FBT001
        """Populates the component recursively from a string."""
        stack = []  # a stack of components
        comps = []
        for line in Contentlines.from_ical(st):  # raw parsing
            if not line:
                continue

            try:
                name, params, vals = line.parts()
            except ValueError as e:
                # if unable to parse a line within a component
                # that ignores exceptions, mark the component
                # as broken and skip the line. otherwise raise.
                component = stack[-1] if stack else None
                if not component or not component.ignore_exceptions:
                    raise
                component.errors.append((None, str(e)))
                continue

            uname = name.upper()
            # check for start of component
            if uname == "BEGIN":
                # try and create one of the components defined in the spec,
                # otherwise get a general Components for robustness.
                c_name = vals.upper()
                c_class = cls.get_component_class(c_name)
                # If component factory cannot resolve ``c_name``, the generic
                # ``Component`` class is used which does not have the name set.
                # That's opposed to the usage of ``cls``, which represents a
                # more concrete subclass with a name set (e.g. VCALENDAR).
                component = c_class()
                if not getattr(component, "name", ""):  # undefined components
                    component.name = c_name
                stack.append(component)
            # check for end of event
            elif uname == "END":
                # we are done adding properties to this component
                # so pop it from the stack and add it to the new top.
                if not stack:
                    # The stack is currently empty, the input must be invalid
                    raise ValueError("END encountered without an accompanying BEGIN!")

                component = stack.pop()
                if not stack:  # we are at the end
                    comps.append(component)
                else:
                    stack[-1].add_component(component)
                if vals == "VTIMEZONE" and "TZID" in component:
                    tzp.cache_timezone_component(component)
            # we are adding properties to the current top of the stack
            else:
                # Extract VALUE parameter if present
                value_param = params.get("VALUE") if params else None
                factory = cls.types_factory.for_property(name, value_param)
                component = stack[-1] if stack else None
                if not component:
                    # only accept X-COMMENT at the end of the .ics file
                    # ignore these components in parsing
                    if uname == "X-COMMENT":
                        break
                    raise ValueError(
                        f'Property "{name}" does not have a parent component.'
                    )
                datetime_names = (
                    "DTSTART",
                    "DTEND",
                    "RECURRENCE-ID",
                    "DUE",
                    "RDATE",
                    "EXDATE",
                )
                try:
                    if name == "FREEBUSY":
                        vals = vals.split(",")
                        if "TZID" in params:
                            parsed_components = [
                                factory(factory.from_ical(val, params["TZID"]))
                                for val in vals
                            ]
                        else:
                            parsed_components = [
                                factory(factory.from_ical(val)) for val in vals
                            ]
                    elif name in datetime_names and "TZID" in params:
                        parsed_components = [
                            factory(factory.from_ical(vals, params["TZID"]))
                        ]
                    # Workaround broken ICS files with empty RDATE
                    elif name == "RDATE" and vals == "":
                        parsed_components = []
                    else:
                        parsed_components = [factory(factory.from_ical(vals))]
                except ValueError as e:
                    if not component.ignore_exceptions:
                        raise
                    component.errors.append((uname, str(e)))
                else:
                    for parsed_component in parsed_components:
                        parsed_component.params = params
                        component.add(name, parsed_component, encode=0)

        if multiple:
            return comps
        if len(comps) > 1:
            raise ValueError(
                cls._format_error(
                    "Found multiple components where only one is allowed", st
                )
            )
        if len(comps) < 1:
            raise ValueError(
                cls._format_error(
                    "Found no components where exactly one is required", st
                )
            )
        return comps[0]

    @staticmethod
    def _format_error(error_description, bad_input, elipsis="[...]"):
        # there's three character more in the error, ie. ' ' x2 and a ':'
        max_error_length = 100 - 3
        if len(error_description) + len(bad_input) + len(elipsis) > max_error_length:
            truncate_to = max_error_length - len(error_description) - len(elipsis)
            return f"{error_description}: {bad_input[:truncate_to]} {elipsis}"
        return f"{error_description}: {bad_input}"

    def content_line(self, name, value, sorted: bool = True):  # noqa: A002, FBT001
        """Returns property as content line."""
        params = getattr(value, "params", Parameters())
        return Contentline.from_parts(name, params, value, sorted=sorted)

    def content_lines(self, sorted: bool = True):  # noqa: A002, FBT001
        """Converts the Component and subcomponents into content lines."""
        contentlines = Contentlines()
        for name, value in self.property_items(sorted=sorted):
            cl = self.content_line(name, value, sorted=sorted)
            contentlines.append(cl)
        contentlines.append("")  # remember the empty string in the end
        return contentlines

    def to_ical(self, sorted: bool = True):  # noqa: A002, FBT001
        """
        :param sorted: Whether parameters and properties should be
                       lexicographically sorted.
        """

        content_lines = self.content_lines(sorted=sorted)
        return content_lines.to_ical()

    def __repr__(self):
        """String representation of class with all of it's subcomponents."""
        subs = ", ".join(str(it) for it in self.subcomponents)
        return (
            f"{self.name or type(self).__name__}"
            f"({dict(self)}{', ' + subs if subs else ''})"
        )

    def __eq__(self, other):
        if len(self.subcomponents) != len(other.subcomponents):
            return False

        properties_equal = super().__eq__(other)
        if not properties_equal:
            return False

        # The subcomponents might not be in the same order,
        # neither there's a natural key we can sort the subcomponents by nor
        # are the subcomponent types hashable, so  we cant put them in a set to
        # check for set equivalence. We have to iterate over the subcomponents
        # and look for each of them in the list.
        for subcomponent in self.subcomponents:
            if subcomponent not in other.subcomponents:
                return False

        return True

    DTSTAMP = stamp = single_utc_property(
        "DTSTAMP",
        """RFC 5545:

        Conformance:  This property MUST be included in the "VEVENT",
        "VTODO", "VJOURNAL", or "VFREEBUSY" calendar components.

        Description: In the case of an iCalendar object that specifies a
        "METHOD" property, this property specifies the date and time that
        the instance of the iCalendar object was created.  In the case of
        an iCalendar object that doesn't specify a "METHOD" property, this
        property specifies the date and time that the information
        associated with the calendar component was last revised in the
        calendar store.

        The value MUST be specified in the UTC time format.

        In the case of an iCalendar object that doesn't specify a "METHOD"
        property, this property is equivalent to the "LAST-MODIFIED"
        property.
    """,
    )
    LAST_MODIFIED = single_utc_property(
        "LAST-MODIFIED",
        """RFC 5545:

        Purpose:  This property specifies the date and time that the
        information associated with the calendar component was last
        revised in the calendar store.

        Note: This is analogous to the modification date and time for a
        file in the file system.

        Conformance:  This property can be specified in the "VEVENT",
        "VTODO", "VJOURNAL", or "VTIMEZONE" calendar components.
    """,
    )

    @property
    def last_modified(self) -> datetime:
        """Datetime when the information associated with the component was last revised.

        Since :attr:`LAST_MODIFIED` is an optional property,
        this returns :attr:`DTSTAMP` if :attr:`LAST_MODIFIED` is not set.
        """
        return self.LAST_MODIFIED or self.DTSTAMP

    @last_modified.setter
    def last_modified(self, value):
        self.LAST_MODIFIED = value

    @last_modified.deleter
    def last_modified(self):
        del self.LAST_MODIFIED

    @property
    def created(self) -> datetime:
        """Datetime when the information associated with the component was created.

        Since :attr:`CREATED` is an optional property,
        this returns :attr:`DTSTAMP` if :attr:`CREATED` is not set.
        """
        return self.CREATED or self.DTSTAMP

    @created.setter
    def created(self, value):
        self.CREATED = value

    @created.deleter
    def created(self):
        del self.CREATED

    def is_thunderbird(self) -> bool:
        """Whether this component has attributes that indicate that Mozilla Thunderbird created it."""  # noqa: E501
        return any(attr.startswith("X-MOZ-") for attr in self.keys())

    @staticmethod
    def _utc_now():
        """Return now as UTC value."""
        return datetime.now(timezone.utc)

    uid = uid_property
    comments = comments_property
    links = links_property
    related_to = related_to_property
    concepts = concepts_property
    refids = refids_property

    CREATED = single_utc_property(
        "CREATED",
        """
        CREATED specifies the date and time that the calendar
        information was created by the calendar user agent in the calendar
        store.

        Conformance:
            The property can be specified once in "VEVENT",
            "VTODO", or "VJOURNAL" calendar components.  The value MUST be
            specified as a date with UTC time.

        """,
    )

    _validate_new = True

    @staticmethod
    def _validate_start_and_end(start, end):
        """This validates start and end.

        Raises:
            InvalidCalendar: If the information is not valid
        """
        if start is None or end is None:
            return
        if start > end:
            raise InvalidCalendar("end must be after start")

    @classmethod
    def new(
        cls,
        created: date | None = None,
        comments: list[str] | str | None = None,
        concepts: CONCEPTS_TYPE_SETTER = None,
        last_modified: date | None = None,
        links: LINKS_TYPE_SETTER = None,
        refids: list[str] | str | None = None,
        related_to: RELATED_TO_TYPE_SETTER = None,
        stamp: date | None = None,
    ) -> Component:
        """Create a new component.

        Arguments:
            comments: The :attr:`comments` of the component.
            concepts: The :attr:`concepts` of the component.
            created: The :attr:`created` of the component.
            last_modified: The :attr:`last_modified` of the component.
            links: The :attr:`links` of the component.
            related_to: The :attr:`related_to` of the component.
            stamp: The :attr:`DTSTAMP` of the component.

        Raises:
            InvalidCalendar: If the content is not valid according to :rfc:`5545`.

        .. warning:: As time progresses, we will be stricter with the validation.
        """
        component = cls()
        component.DTSTAMP = stamp
        component.created = created
        component.last_modified = last_modified
        component.comments = comments
        component.links = links
        component.related_to = related_to
        component.concepts = concepts
        component.refids = refids
        return component

    def to_jcal(self) -> list:
        """Convert this component to a jCal object.

        Returns:
            jCal object

        See also :attr:`to_json`.

        In this example, we create a simple VEVENT component and convert it to jCal:

        .. code-block:: pycon

            >>> from icalendar import Event
            >>> from datetime import date
            >>> from pprint import pprint
            >>> event = Event.new(summary="My Event", start=date(2025, 11, 22))
            >>> pprint(event.to_jcal())
            ['vevent',
             [['dtstamp', {}, 'date-time', '2025-05-17T08:06:12Z'],
              ['summary', {}, 'text', 'My Event'],
              ['uid', {}, 'text', 'd755cef5-2311-46ed-a0e1-6733c9e15c63'],
              ['dtstart', {}, 'date', '2025-11-22']],
             []]
        """
        properties = []
        for key, value in self.items():
            for item in value if isinstance(value, list) else [value]:
                properties.append(item.to_jcal(key.lower()))
        return [
            self.name.lower(),
            properties,
            [subcomponent.to_jcal() for subcomponent in self.subcomponents],
        ]

    def to_json(self) -> str:
        """Return this component as a jCal JSON string.

        Returns:
            JSON string

        See also :attr:`to_jcal`.
        """
        return json.dumps(self.to_jcal())

    @classmethod
    def from_jcal(cls, jcal: str | list) -> Component:
        """Create a component from a jCal list.

        Args:
            jcal: jCal list or JSON string according to :rfc:`7265`.

        Raises:
            JCalParsingError: If the jCal provided is invalid.
            ~json.JSONDecodeError: If the provided string is not valid JSON.

        This reverses :func:`to_json` and :func:`to_jcal`.

        The following code parses an example from :rfc:`7265`:

        .. code-block:: pycon

            >>> from icalendar import Component
            >>> jcal = ["vcalendar",
            ...   [
            ...     ["calscale", {}, "text", "GREGORIAN"],
            ...     ["prodid", {}, "text", "-//Example Inc.//Example Calendar//EN"],
            ...     ["version", {}, "text", "2.0"]
            ...   ],
            ...   [
            ...     ["vevent",
            ...       [
            ...         ["dtstamp", {}, "date-time", "2008-02-05T19:12:24Z"],
            ...         ["dtstart", {}, "date", "2008-10-06"],
            ...         ["summary", {}, "text", "Planning meeting"],
            ...         ["uid", {}, "text", "4088E990AD89CB3DBB484909"]
            ...       ],
            ...       []
            ...     ]
            ...   ]
            ... ]
            >>> calendar = Component.from_jcal(jcal)
            >>> print(calendar.name)
            VCALENDAR
            >>> print(calendar.prodid)
            -//Example Inc.//Example Calendar//EN
            >>> event = calendar.events[0]
            >>> print(event.summary)
            Planning meeting

        """
        if isinstance(jcal, str):
            jcal = json.loads(jcal)
        if not isinstance(jcal, list) or len(jcal) != 3:
            raise JCalParsingError(
                "A component must be a list with 3 items.", cls, value=jcal
            )
        name, properties, subcomponents = jcal
        if not isinstance(name, str):
            raise JCalParsingError(
                "The name must be a string.", cls, path=[0], value=name
            )
        if name.upper() != cls.name:
            # delegate to correct component class
            component_cls = cls.get_component_class(name.upper())
            return component_cls.from_jcal(jcal)
        component = cls()
        if not isinstance(properties, list):
            raise JCalParsingError(
                "The properties must be a list.", cls, path=1, value=properties
            )
        for i, prop in enumerate(properties):
            JCalParsingError.validate_property(prop, cls, path=[1, i])
            prop_name = prop[0]
            prop_value = prop[2]
            prop_cls: type[VPROPERTY] = cls.types_factory.for_property(
                prop_name, prop_value
            )
            with JCalParsingError.reraise_with_path_added(1, i):
                v_prop = prop_cls.from_jcal(prop)
            # if we use the default value for that property, we can delete the
            # VALUE parameter
            if prop_cls == cls.types_factory.for_property(prop_name):
                del v_prop.VALUE
            component.add(prop_name, v_prop)
        if not isinstance(subcomponents, list):
            raise JCalParsingError(
                "The subcomponents must be a list.", cls, 2, value=subcomponents
            )
        for i, subcomponent in enumerate(subcomponents):
            with JCalParsingError.reraise_with_path_added(2, i):
                component.subcomponents.append(cls.from_jcal(subcomponent))
        return component


__all__ = ["Component"]



"""

See https://github.com/collective/icalendar/issues/322
"""

from icalendar.cal.calendar import Calendar
from icalendar.cal.event import Event


def test_issue_322_single_string_split_into_multiple_categories(calendars):
    calendar = Calendar()
    event = Event()
    event.add("summary", "Event with bare string as argument for categories")
    event.add("categories", "Lecture")
    calendar.add_component(event)
    assert calendar.to_ical() == calendars.issue_322_expected_calendar.raw_ics



import unittest

from icalendar.parser_tools import data_encode, to_unicode, from_unicode


class TestParserTools(unittest.TestCase):
    def test_parser_tools_to_unicode(self):
        self.assertEqual(to_unicode(b"spam"), "spam")
        self.assertEqual(to_unicode("spam"), "spam")
        self.assertEqual(to_unicode(b"spam"), "spam")
        self.assertEqual(to_unicode(b"\xc6\xb5"), "\u01b5")
        self.assertEqual(to_unicode(b"\xc6\xb5"), "\u01b5")
        self.assertEqual(to_unicode(b"\xc6\xb5", encoding="ascii"), "\u01b5")
        self.assertEqual(to_unicode(1), 1)
        self.assertIsNone(to_unicode(None))

    def test_parser_tools_from_unicode(self):
        self.assertEqual(from_unicode("\u01b5", encoding="ascii"), b"\xc6\xb5")

    def test_parser_tools_data_encode(self):
        data1 = {
            "k1": "v1",
            "k2": "v2",
            "k3": "v3",
            "li1": ["it1", "it2", {"k4": "v4", "k5": "v5"}, 123],
        }
        res = {
            b"k3": b"v3",
            b"k2": b"v2",
            b"k1": b"v1",
            b"li1": [b"it1", b"it2", {b"k5": b"v5", b"k4": b"v4"}, 123],
        }
        self.assertEqual(data_encode(data1), res)



"""Test the new property types UID and XML-REFERENCE.

- creation
- parameters
- serialization
- deserialization
- attributes
"""

import pytest

from icalendar import Calendar, Event, vUid, vUri, vXmlReference


@pytest.fixture(
    params=[
        "123e4567-e89b-12d3-a456-426614174000",
        "550e8400-e29b-41d4-a716-446655440000",
    ]
)
def uid(request):
    return request.param


@pytest.fixture(
    params=[
        "https://example.com/events",
        "https://example.com/tasks/01234567-abcd1234.ics",
    ]
)
def uri(request):
    return request.param


@pytest.fixture(
    params=[
        (
            "https://example.com/xmlDocs/bidFramework.xml#xpointer(descendant::CostStruc/range-to(following::CostStrucEND[1]))",
            "descendant::CostStruc/range-to(following::CostStrucEND[1])",
        ),
        (
            "https://www.w3.org/TR/2003/REC-xptr-framework-20030325/#xpointer(id('r%C3%A9sum%C3%A9'))",
            "id('résumé')",
        ),
    ]
)
def xml_reference(request):
    return request.param


def test_uid_creation(uid):
    """Test creation of UID property."""
    uid_prop = vUid(uid)
    assert repr(uid_prop) == f"vUid({uid!r})"
    assert uid_prop.ical_value == uid
    assert uid_prop.uid == uid
    assert_default_parameters(uid_prop)


def test_uid_new(test_uid):
    """The new uid should be tested."""
    uid_prop = vUid.new()
    assert uid_prop.uid == test_uid
    assert_default_parameters(uid_prop)
    assert uid_prop.VALUE == "UID"


def test_vUri_creation(uri):
    """Test creation of a URI property."""
    uri_prop = vUri(uri)
    assert str(uri_prop) == uri
    assert uri_prop.ical_value == uri
    assert uri_prop.uri == uri
    assert_default_parameters(uri_prop)
    assert uri_prop.VALUE == "URI"


def test_vXmlReference_creation(xml_reference):
    """Test creation of a URI property."""
    uri, x_pointer = xml_reference
    xml_prop = vXmlReference(uri)
    assert str(xml_prop) == uri
    assert xml_prop.ical_value == uri
    assert xml_prop.uri == uri
    assert_default_parameters(xml_prop)
    assert xml_prop.x_pointer == x_pointer
    assert xml_prop.VALUE == "XML-REFERENCE"


def assert_default_parameters(prop):
    """Test an empty property."""
    assert prop.LABEL is None
    assert prop.LANGUAGE is None
    assert prop.FMTTYPE is None
    assert prop.LINKREL is None
    assert repr(prop).startswith(prop.__class__.__name__ + "(")
    assert repr(prop).endswith(")")
    assert prop.params.value is None
    assert prop.params.get("VALUE") is None


@pytest.mark.parametrize("no_x_pointer", ["", "asd)", "xpointer(id('r%C3%A9sum%C3%A9'"])
def test_not_x_pointer(no_x_pointer):
    """Test that the x_pointer throws a value error."""
    xml = vXmlReference("http://asd#" + no_x_pointer)
    assert xml.x_pointer is None


mark_valid_values = pytest.mark.parametrize(
    ("name", "value"),
    [
        ("LABEL", None),
        ("LABEL", "Venue"),
        ("LANGUAGE", None),
        ("LANGUAGE", "de"),
        ("LANGUAGE", "en-gb"),
        ("LINKREL", None),
        ("LINKREL", "SOURCE"),
        ("LINKREL", "https://example.com/linkrel/derivedFrom"),
        ("LINKREL", "icon"),
        ("FMTTYPE", None),
        ("FMTTYPE", "text/html"),
    ],
)


def vUri_example():
    """New example instance."""
    return vUri("http://example.com")


def vXmlReference_example():
    """New example instance."""
    return vXmlReference("http://example.com")


mark_new_prop = pytest.mark.parametrize(
    "new_prop", [vUri_example, vUid.new, vXmlReference_example]
)


@mark_new_prop
@mark_valid_values
def test_set_attributes(new_prop, name, value):
    """Set the values and test"""
    prop = new_prop()
    assert hasattr(prop, name)
    setattr(prop, name, value)
    print(prop.params)
    assert getattr(prop, name) == value
    if value is None:
        assert name not in prop.params
    else:
        assert prop.params[name] == value


@mark_new_prop
@mark_valid_values
def test_delete_attribute_setting_to_none(new_prop, name, value):
    """The attribute can be deleted by setting to None."""
    prop = new_prop()
    setattr(prop, name, value)
    setattr(prop, name, None)
    assert name not in prop.params
    assert getattr(prop, name) is None


@mark_new_prop
@mark_valid_values
def test_delete_attribute(new_prop, name, value):
    """The attribute can be deleted."""
    prop = new_prop()
    setattr(prop, name, value)
    delattr(prop, name)
    assert name not in prop.params
    assert getattr(prop, name) is None


def test_parse_example_calendar(calendars):
    """We should make sure that examples from the RFC work."""
    calendar: Calendar = calendars.rfc_9253_examples
    event = calendar.events[0]
    links = event.links
    assert len(links) == 3
    assert links[0].uri == "https://example.com/events"
    assert links[0].LINKREL == "SOURCE"
    assert links[0].FMTTYPE is None
    assert links[0].LABEL == "Venue"
    assert isinstance(links[0], vUri)

    assert links[1].uri == "https://example.com/tasks/01234567-abcd1234.ics"
    assert links[1].LINKREL == "https://example.com/linkrel/derivedFrom"
    assert links[1].FMTTYPE is None
    assert links[1].LABEL is None
    assert isinstance(links[1], vUri)

    assert (
        links[2].xml_reference
        == "https://example.com/xmlDocs/bidFramework.xml#xpointer(descendant::CostStruc/range-to(following::CostStrucEND[1]))"
    )
    assert links[2].LINKREL == "https://example.com/linkrel/costStructure"
    assert links[2].FMTTYPE is None
    assert links[2].LABEL is None
    assert isinstance(links[2], vXmlReference)


def test_parse_uid_example(calendars):
    """Parse the UID type."""
    calendar: Calendar = calendars.rfc_9253_examples
    event = calendar.events[1]
    links = event.links
    assert len(links) == 1
    assert links[0].uid == "links-rfc-9253-section-8.2"
    assert links[0].LINKREL == "REFERENCE"
    assert links[0].FMTTYPE is None
    assert links[0].LABEL is None
    assert isinstance(links[0], vUid)


def test_linkrel_is_always_quoted():
    """Test that the LINKREL is always quoted."""
    link = vUri("https://example.com", params={"LINKREL": "SOURCE"})
    event = Event()
    event.links = [link]
    ical = event.to_ical().decode()
    assert 'LINKREL="SOURCE"' in ical



"""Use zoneinfo timezones"""

from __future__ import annotations

import copy
import copyreg
import functools
from datetime import datetime, tzinfo
from io import StringIO
from typing import TYPE_CHECKING, Optional

from dateutil.rrule import rrule, rruleset
from dateutil.tz import tzical
from dateutil.tz.tz import _tzicalvtz

from icalendar.compatibility import zoneinfo
from icalendar.tools import is_date, to_datetime

from .provider import TZProvider

if TYPE_CHECKING:
    from icalendar import prop
    from icalendar.cal import Timezone
    from icalendar.prop import vDDDTypes


class ZONEINFO(TZProvider):
    """Provide icalendar with timezones from zoneinfo."""

    name = "zoneinfo"
    utc = zoneinfo.ZoneInfo("UTC")
    _available_timezones = zoneinfo.available_timezones()

    def localize(self, dt: datetime, tz: zoneinfo.ZoneInfo) -> datetime:
        """Localize a datetime to a timezone."""
        return dt.replace(tzinfo=tz)

    def localize_utc(self, dt: datetime) -> datetime:
        """Return the datetime in UTC."""
        if getattr(dt, "tzinfo", False) and dt.tzinfo is not None:
            return dt.astimezone(self.utc)
        return self.localize(dt, self.utc)

    def timezone(self, name: str) -> Optional[tzinfo]:
        """Return a timezone with a name or None if we cannot find it."""
        try:
            return zoneinfo.ZoneInfo(name)
        except zoneinfo.ZoneInfoNotFoundError:
            pass
        except ValueError:
            # ValueError: ZoneInfo keys may not be absolute paths, got: /Europe/CUSTOM
            pass

    def knows_timezone_id(self, tzid: str) -> bool:
        """Whether the timezone is already cached by the implementation."""
        return tzid in self._available_timezones

    def fix_rrule_until(self, rrule: rrule, ical_rrule: prop.vRecur) -> None:
        """Make sure the until value works for the rrule generated from the ical_rrule."""  # noqa: E501
        if not {"UNTIL", "COUNT"}.intersection(ical_rrule.keys()):
            # zoninfo does not know any transition dates after 2038
            rrule._until = datetime(2038, 12, 31, tzinfo=self.utc)  # noqa: SLF001

    def create_timezone(self, tz: Timezone.Timezone) -> tzinfo:
        """Create a timezone from the given information."""
        try:
            return self._create_timezone(tz)
        except ValueError:
            # We might have a custom component in there.
            # see https://github.com/python/cpython/issues/120217
            tz = copy.deepcopy(tz)
            for sub in tz.walk():
                for attr in list(sub.keys()):
                    if attr.lower().startswith("x-"):
                        sub.pop(attr)
            for sub in tz.subcomponents:
                start: vDDDTypes = sub.get("DTSTART")
                if start and is_date(start.dt):
                    # ValueError: Unsupported DTSTART param in VTIMEZONE: VALUE=DATE
                    sub.DTSTART = to_datetime(start.dt)
            return self._create_timezone(tz)

    def _create_timezone(self, tz: Timezone.Timezone) -> tzinfo:
        """Create a timezone and maybe fail"""
        file = StringIO(tz.to_ical().decode("UTF-8", "replace"))
        return tzical(file).get()

    def uses_pytz(self) -> bool:
        """Whether we use pytz."""
        return False

    def uses_zoneinfo(self) -> bool:
        """Whether we use zoneinfo."""
        return True


def pickle_tzicalvtz(tzicalvtz: _tzicalvtz):
    """Because we use dateutil.tzical, we need to make it pickle-able."""
    return _tzicalvtz, (tzicalvtz._tzid, tzicalvtz._comps)  # noqa: SLF001


copyreg.pickle(_tzicalvtz, pickle_tzicalvtz)


def pickle_rrule_with_cache(self: rrule):
    """Make sure we can also pickle rrules that cache.

    This is mainly copied from rrule.replace.
    """
    new_kwargs = {
        "interval": self._interval,
        "count": self._count,
        "dtstart": self._dtstart,
        "freq": self._freq,
        "until": self._until,
        "wkst": self._wkst,
        "cache": self._cache is not None,
    }
    new_kwargs.update(self._original_rule)
    # from https://stackoverflow.com/a/64915638/1320237
    return functools.partial(rrule, new_kwargs.pop("freq"), **new_kwargs), ()


copyreg.pickle(rrule, pickle_rrule_with_cache)


def pickle_rruleset_with_cache(rs: rruleset):
    """Pickle an rruleset."""
    return unpickle_rruleset_with_cache, (
        rs._rrule,  # noqa: SLF001
        rs._rdate,  # noqa: SLF001
        rs._exrule,  # noqa: SLF001
        rs._exdate,  # noqa: SLF001
        rs._cache is not None,  # noqa: SLF001
    )


def unpickle_rruleset_with_cache(rrule, rdate, exrule, exdate, cache):
    """unpickling the rruleset."""
    rs = rruleset(cache)
    for o in rrule:
        rs.rrule(o)
    for o in rdate:
        rs.rdate(o)
    for o in exrule:
        rs.exrule(o)
    for o in exdate:
        rs.exdate(o)
    return rs


copyreg.pickle(rruleset, pickle_rruleset_with_cache)

__all__ = ["ZONEINFO"]



"""These are tests for Issue #557

TL;DR: Component._encode lost given parameters
if the object to encode was already of native type,
making its behavior unexpected.

see https://github.com/collective/icalendar/issues/557
"""

import unittest

from icalendar.cal.component import Component


class TestComponentEncode(unittest.TestCase):
    def test_encode_non_native_parameters(self):
        """Test _encode to add parameters to non-natives"""
        self.__assert_native_content(self.summary)
        self.__assert_native_kept_parameters(self.summary)

    def test_encode_native_keep_params_None(self):
        """_encode should keep parameters on natives
        if parameters=None
        """
        new_sum = self.__add_params(
            self.summary,
            parameters=None,
        )
        self.__assert_native_content(new_sum)
        self.__assert_native_kept_parameters(new_sum)

    def test_encode_native_keep_params_empty(self):
        """_encode should keep paramters on natives
        if parameters={}
        """
        new_sum = self.__add_params(
            self.summary,
            parameters={},
        )
        self.__assert_native_content(new_sum)
        self.__assert_native_kept_parameters(new_sum)

    def test_encode_native_append_params(self):
        """_encode should append paramters on natives
        keeping old parameters
        """
        new_sum = self.__add_params(
            self.summary,
            parameters={"X-PARAM": "Test123"},
        )
        self.__assert_native_content(new_sum)
        self.__assert_native_kept_parameters(new_sum)
        self.assertParameter(new_sum, "X-PARAM", "Test123")

    def test_encode_native_overwrite_params(self):
        """_encode should overwrite single parameters
        if they have the same name as old ones"""
        new_sum = self.__add_params(
            self.summary,
            parameters={"LANGUAGE": "de"},
        )
        self.__assert_native_content(new_sum)
        self.assertParameter(new_sum, "LANGUAGE", "de")

    def test_encode_native_remove_params(self):
        """_encode should remove single parameters
        if they are explicitly set to None"""
        new_sum = self.__add_params(
            self.summary,
            parameters={"LANGUAGE": None},
        )
        self.__assert_native_content(new_sum)
        self.assertParameterMissing(new_sum, "LANGUAGE")

    def test_encode_native_remove_already_missing(self):
        """_encode should ignore removing a parameter
        that was already missing"""
        self.assertParameterMissing(self.summary, "X-MISSING")
        new_sum = self.__add_params(
            self.summary,
            parameters={"X-MISSING": None},
        )
        self.__assert_native_content(new_sum)
        self.__assert_native_kept_parameters(new_sum)
        self.assertParameterMissing(self.summary, "X-MISSING")

    def test_encode_native_full_test(self):
        """full test case with keeping, overwriting & removing properties"""
        # preperation
        orig_sum = self.__add_params(
            self.summary,
            parameters={
                "X-OVERWRITE": "overwrite me!",
                "X-REMOVE": "remove me!",
                "X-MISSING": None,
            },
        )
        # preperation check
        self.__assert_native_content(orig_sum)
        self.__assert_native_kept_parameters(orig_sum)
        self.assertParameter(orig_sum, "X-OVERWRITE", "overwrite me!")
        self.assertParameter(orig_sum, "X-REMOVE", "remove me!")
        self.assertParameterMissing(orig_sum, "X-MISSING")
        # modification
        new_sum = self.__add_params(
            orig_sum,
            parameters={
                "X-OVERWRITE": "overwritten",
                "X-REMOVE": None,
                "X-MISSING": None,
            },
        )
        # final asserts
        self.__assert_native_content(new_sum)
        self.__assert_native_kept_parameters(new_sum)
        self.assertParameter(new_sum, "X-OVERWRITE", "overwritten")
        self.assertParameterMissing(new_sum, "X-REMOVE")
        self.assertParameterMissing(new_sum, "X-MISSING")

    def setUp(self):
        self.summary = self.__gen_native()

    def __assert_native_kept_parameters(self, obj):
        self.assertParameter(obj, "LANGUAGE", "en")

    def __assert_native_content(self, obj):
        assert obj == "English Summary"

    def __add_params(self, obj, parameters):
        return Component._encode(
            "SUMMARY",
            obj,
            parameters=parameters,
            encode=True,
        )

    def __gen_native(self):
        return Component._encode(
            "SUMMARY",
            "English Summary",
            parameters={
                "LANGUAGE": "en",
            },
            encode=True,
        )

    def assertParameterMissing(self, obj, name):
        assert name not in obj.params

    def assertParameter(self, obj, name, val):
        assert name in obj.params
        assert obj.params[name] == val



import unittest

import pytest

from icalendar.parser import (
    Contentline,
    Contentlines,
    Parameters,
    dquote,
    foldline,
    q_join,
    q_split,
)
from icalendar.prop import vText


class IcalendarTestCase(unittest.TestCase):
    def setUp(self):
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_long_lines(self):
        c = Contentlines([Contentline("BEGIN:VEVENT")])
        c.append(Contentline("".join("123456789 " * 10)))
        assert (
            c.to_ical()
            == b"BEGIN:VEVENT\r\n123456789 123456789 123456789 123456789 "  # noqa: ISC003
            + b"123456789 123456789 123456789 1234\r\n 56789 123456789 "
            + b"123456789 \r\n"
        )

        # from doctests
        # Notice that there is an extra empty string in the end of the content
        # lines. That is so they can be easily joined with:
        # '\r\n'.join(contentlines))
        assert Contentlines.from_ical("A short line\r\n") == ["A short line", ""]
        assert Contentlines.from_ical("A faked\r\n  long line\r\n") == [
            "A faked long line",
            "",
        ]
        assert Contentlines.from_ical(
            "A faked\r\n  long line\r\nAnd another lin\r\n\te that is folded\r\n"
        ) == ["A faked long line", "And another line that is folded", ""]

    def test_contentline_class(self):
        assert (
            Contentline("Si meliora dies, ut vina, poemata reddit").to_ical()
            == b"Si meliora dies, ut vina, poemata reddit"
        )

        # A long line gets folded
        c = Contentline("".join(["123456789 "] * 10)).to_ical()
        assert (
            c
            == b"123456789 123456789 123456789 123456789 123456789 123456789 "  # noqa: ISC003
            + b"123456789 1234\r\n 56789 123456789 123456789 "
        )

        # A folded line gets unfolded
        assert (
            Contentline.from_ical(c)
            == "123456789 123456789 123456789 123456789 123456789 123456789 "  # noqa: ISC003
            + "123456789 123456789 123456789 123456789 "
        )

        # https://tools.ietf.org/html/rfc5545#section-3.3.11
        # An intentional formatted text line break MUST only be included in
        # a "TEXT" property value by representing the line break with the
        # character sequence of BACKSLASH, followed by a LATIN SMALL LETTER
        # N or a LATIN CAPITAL LETTER N, that is "\n" or "\N".

        # Newlines are not allowed in content lines
        self.assertRaises(AssertionError, Contentline, b"1234\r\n\r\n1234")

        assert Contentline("1234\\n\\n1234").to_ical() == b"1234\\n\\n1234"

        # We do not fold within a UTF-8 character
        c = Contentline(
            b"This line has a UTF-8 character where it should be "
            b"folded. Make sure it g\xc3\xabts folded before that "
            b"character."
        )

        assert b"\xc3\xab" in c.to_ical()

        # Another test of the above
        c = Contentline(b"x" * 73 + b"\xc3\xab" + b"\\n " + b"y" * 10)

        assert c.to_ical().count(b"\xc3") == 1

        # Don't fail if we fold a line that is exactly X times 74 characters
        # long
        Contentline("".join(["x"] * 148)).to_ical()

        # It can parse itself into parts,
        # which is a tuple of (name, params, vals)
        assert Contentline("dtstart:20050101T120000").parts() == (
            "dtstart",
            Parameters({}),
            "20050101T120000",
        )

        assert Contentline("dtstart;value=datetime:20050101T120000").parts() == (
            "dtstart",
            Parameters({"VALUE": "datetime"}),
            "20050101T120000",
        )

        c = Contentline(
            "ATTENDEE;CN=Max Rasmussen;ROLE=REQ-PARTICIPANT:MAILTO:maxm@example.com"
        )
        assert c.parts() == (
            "ATTENDEE",
            Parameters({"ROLE": "REQ-PARTICIPANT", "CN": "Max Rasmussen"}),
            "MAILTO:maxm@example.com",
        )
        assert (
            c.to_ical().decode("utf-8")
            == "ATTENDEE;CN=Max Rasmussen;ROLE=REQ-PARTICIPANT:MAILTO:maxm@example.com"
        )

        # and back again
        # NOTE: we are quoting property values with spaces in it.
        parts = (
            "ATTENDEE",
            Parameters({"ROLE": "REQ-PARTICIPANT", "CN": "Max Rasmussen"}),
            "MAILTO:maxm@example.com",
        )
        assert (
            Contentline.from_parts(*parts)
            == 'ATTENDEE;CN="Max Rasmussen";ROLE=REQ-PARTICIPANT:MAILTO:maxm@example.com'
        )

        # and again
        parts = ("ATTENDEE", Parameters(), "MAILTO:maxm@example.com")
        assert Contentline.from_parts(*parts) == "ATTENDEE:MAILTO:maxm@example.com"

        # A value can also be any of the types defined in PropertyValues
        parts = ("ATTENDEE", Parameters(), vText("MAILTO:test@example.com"))
        assert Contentline.from_parts(*parts) == "ATTENDEE:MAILTO:test@example.com"

        # A value in UTF-8
        parts = ("SUMMARY", Parameters(), vText("INternational char æ ø å"))
        assert Contentline.from_parts(*parts) == "SUMMARY:INternational char æ ø å"

        # A value can also be unicode
        parts = ("SUMMARY", Parameters(), vText("INternational char æ ø å"))
        assert Contentline.from_parts(*parts) == "SUMMARY:INternational char æ ø å"

        # Traversing could look like this.
        name, params, vals = c.parts()
        assert name == "ATTENDEE"
        assert vals == "MAILTO:maxm@example.com"
        assert sorted(params.items()) == sorted(
            [("ROLE", "REQ-PARTICIPANT"), ("CN", "Max Rasmussen")]
        )

        # And the traditional failure
        with pytest.raises(
            ValueError, match="Content line could not be parsed into parts"
        ):
            Contentline("ATTENDEE;maxm@example.com").parts()

        # Another failure:
        with pytest.raises(
            ValueError, match="Content line could not be parsed into parts"
        ):
            Contentline(":maxm@example.com").parts()

        assert Contentline("key;param=:value").parts() == (
            "key",
            Parameters({"PARAM": ""}),
            "value",
        )

        assert Contentline('key;param="pvalue":value').parts() == (
            "key",
            Parameters({"PARAM": "pvalue"}),
            "value",
        )

        # Should bomb on missing param:
        with pytest.raises(
            ValueError, match="Content line could not be parsed into parts"
        ):
            Contentline.from_ical("k;:no param").parts()

        assert Contentline("key;param=pvalue:value", strict=False).parts() == (
            "key",
            Parameters({"PARAM": "pvalue"}),
            "value",
        )

        # If strict is set to True, uppercase param values that are not
        # double-quoted, this is because the spec says non-quoted params are
        # case-insensitive.
        assert Contentline("key;param=pvalue:value", strict=True).parts() == (
            "key",
            Parameters({"PARAM": "PVALUE"}),
            "value",
        )

        assert Contentline('key;param="pValue":value', strict=True).parts() == (
            "key",
            Parameters({"PARAM": "pValue"}),
            "value",
        )

        contains_base64 = (
            b"X-APPLE-STRUCTURED-LOCATION;"
            b'VALUE=URI;X-ADDRESS="Kaiserliche Hofburg, 1010 Wien";'
            b"X-APPLE-MAPKIT-HANDLE=CAESxQEZgr3QZXJyZWljaA==;"
            b"X-APPLE-RADIUS=328.7978217977285;X-APPLE-REFERENCEFRAME=1;"
            b"X-TITLE=Heldenplatz:geo:48.206686,16.363235"
        )

        assert Contentline(contains_base64, strict=True).parts() == (
            "X-APPLE-STRUCTURED-LOCATION",
            Parameters(
                {
                    "X-APPLE-RADIUS": "328.7978217977285",
                    "X-ADDRESS": "Kaiserliche Hofburg, 1010 Wien",
                    "X-APPLE-REFERENCEFRAME": "1",
                    "X-TITLE": "HELDENPLATZ",
                    "X-APPLE-MAPKIT-HANDLE": "CAESXQEZGR3QZXJYZWLJAA==",
                    "VALUE": "URI",
                }
            ),
            "geo:48.206686,16.363235",
        )

    def test_fold_line(self):
        assert foldline("foo") == "foo"
        assert (
            foldline(
                "Lorem ipsum dolor sit amet, consectetur adipiscing "
                "elit. Vestibulum convallis imperdiet dui posuere.",
            )
            == "Lorem ipsum dolor sit amet, consectetur adipiscing elit. "
            "Vestibulum conval\r\n lis imperdiet dui posuere."
        )

        # I don't really get this test
        # at least just but bytes in there
        # porting it to "run" under python 2 & 3 makes it not much better
        with pytest.raises(AssertionError):
            foldline("привет".encode(), limit=3)

        assert foldline("foobar", limit=4) == "foo\r\n bar"
        assert (
            foldline(
                "Lorem ipsum dolor sit amet, consectetur adipiscing elit"
                ". Vestibulum convallis imperdiet dui posuere.",
            )
            == "Lorem ipsum dolor sit amet, consectetur adipiscing elit."  # noqa: ISC003
            + " Vestibulum conval\r\n lis imperdiet dui posuere."
        )
        assert (
            foldline("DESCRIPTION:АБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЬЫЪЭЮЯ")
            == "DESCRIPTION:АБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЬЫЪЭ\r\n ЮЯ"
        )

    def test_value_double_quoting(self):
        assert dquote("Max") == "Max"
        assert dquote("Rasmussen, Max") == '"Rasmussen, Max"'
        assert dquote("name:value") == '"name:value"'

    def test_q_split(self):
        assert q_split('Max,Moller,"Rasmussen, Max"') == [
            "Max",
            "Moller",
            '"Rasmussen, Max"',
        ]

    def test_q_split_bin(self):
        for s in ("X-SOMETHING=ABCDE==", ",,,"):
            for maxsplit in range(-1, 3):
                assert q_split(s, "=", maxsplit=maxsplit) == s.split("=", maxsplit)

    def test_q_join(self):
        assert (
            q_join(["Max", "Moller", "Rasmussen, Max"]) == 'Max,Moller,"Rasmussen, Max"'
        )



"""
Property parameters are represented as a JSON object where each key-
value pair represents the iCalendar parameter name and its value.
The name of the parameter MUST be in lowercase; the original case of
the parameter value MUST be preserved.
"""

from typing import TYPE_CHECKING

import pytest

from icalendar import PARTSTAT, ROLE, VPROPERTY, Event, vCalAddress
from icalendar.parser import Parameters

if TYPE_CHECKING:
    from icalendar.cal.calendar import Calendar


def test_vCalAddressExample():
    """Test the example from vCalAddress."""
    addr = vCalAddress.new(
        "jsmith@example.org",
        partstat=PARTSTAT.ACCEPTED,
        rsvp=True,
        role=ROLE.REQ_PARTICIPANT,
    )
    event = Event()
    event.attendees = [addr]
    event.summary = "Meeting"
    jcal = event.to_jcal()
    assert jcal == [
        "vevent",
        [
            [
                "attendee",
                {"partstat": "ACCEPTED", "rsvp": "TRUE", "role": "REQ-PARTICIPANT"},
                "cal-address",
                "mailto:jsmith@example.org",
            ],
            ["summary", {}, "text", "Meeting"],
        ],
        [],
    ]


def test_parameter_conversion_is_handed_over_to_parameter(v_prop: VPROPERTY, mock):
    """The parameters are part of the jCal conversion."""
    example = v_prop.examples()[0]
    assert isinstance(example.params, Parameters)
    example.params = mock
    return_value = example.params.to_jcal.return_value
    jcal = example.to_jcal("X-NAME")
    mock.to_jcal.assert_called_once()
    assert len(jcal) >= 4
    assert jcal[1] == return_value
    assert jcal[0] == "X-NAME"


def test_all_default_values_are_lowercase(v_prop_example):
    """All the default values should be lowercase.

    :rfc:`7265`:
        If the property has no "VALUE" parameter but has a default value
        type, the default value type is used.
    """
    jcal = v_prop_example.to_jcal("X-NAME")
    ical_type: str = jcal[2]
    assert ical_type.islower()


def test_all_set_values_are_preserved(v_prop_example):
    """Check the set value type.

    :rfc:`5545`:
        All names of [...] property parameter values are case-insensitive.

    :rfc:`7265`:
        If the property has a "VALUE" parameter, that parameter's value
        is used as the value type.

    We can convert them to lowercase but it is not stated explicitely.
    """
    v_prop_example.VALUE = "X-OTHER-VALUE"
    jcal = v_prop_example.to_jcal("X-NAME")
    ical_type: str = jcal[2]
    assert ical_type == "x-other-value"


def test_unknown_value(calendars):
    """Test unknown value.

    :rfc:`7265`:
        If the property has no "VALUE" parameter and has no default value
        type, "unknown" is used.

    """
    calendar: Calendar = calendars.x_location
    jcal = calendar.to_jcal()
    props = jcal[1]
    checked = 0
    for prop in props:
        print(prop)
        if prop[0].startswith("x-"):
            assert prop[2] == "unknown"
            checked += 1
    assert checked >= 3, "We checked at least some properties."


@pytest.mark.parametrize(
    ("event_index", "parameter_index", "expected_value"),
    [
        (
            0,
            0,
            [
                "attendee",
                {
                    "delegated-to": [
                        "mailto:jdoe@example.com",
                        "mailto:jqpublic@example.com",
                    ]
                },
                "cal-address",
                "mailto:jsmith@example.com",
            ],
        ),
        (
            0,
            1,
            [
                "attendee",
                {
                    "delegated-from": [
                        "mailto:jsmith@example.com",
                        "mailto:jdoe@example.com",
                    ]
                },
                "cal-address",
                "mailto:jdoe@example.com",
            ],
        ),
        (
            0,
            2,
            [
                "attendee",
                {
                    "member": [
                        "mailto:projectA@example.com",
                        "mailto:projectB@example.com",
                    ]
                },
                "cal-address",
                "mailto:janedoe@example.com",
            ],
        ),
        (
            1,
            0,
            [
                "attendee",
                {"delegated-to": "mailto:jdoe@example.com"},
                "cal-address",
                "mailto:jsmith@example.com",
            ],
        ),
        (
            1,
            1,
            [
                "attendee",
                {"delegated-from": "mailto:jsmith@example.com"},
                "cal-address",
                "mailto:jdoe@example.com",
            ],
        ),
        (
            1,
            2,
            [
                "attendee",
                {"member": "mailto:projectA@example.com"},
                "cal-address",
                "mailto:janedoe@example.com",
            ],
        ),
    ],
)
def test_parameters_with_values_as_list(
    calendars, event_index, parameter_index, expected_value
):
    """Check the conversion of list value parameters.

    In [RFC5545], some parameters allow using a COMMA-separated list of
    values.  To ease processing in jCal, the value of such parameters
    MUST be represented in an array containing the separated values.
    """
    calendar: Calendar = calendars.rfc_7256_multi_value_parameters
    event = calendar.events[event_index]
    parameter = event.to_jcal()[1][parameter_index]
    assert parameter == expected_value



"""Test that composed values are properly converted."""

from datetime import datetime

from icalendar.cal.event import Event


def test_vDDDLists_timezone(tzp):
    """Test vDDDLists with timezone information."""
    vevent = Event()
    dt1 = tzp.localize(datetime(2013, 1, 1), "Europe/Vienna")
    dt2 = tzp.localize(datetime(2013, 1, 2), "Europe/Vienna")
    dt3 = tzp.localize(datetime(2013, 1, 3), "Europe/Vienna")
    vevent.add("rdate", [dt1, dt2])
    vevent.add("exdate", dt3)
    ical = vevent.to_ical()

    assert b"RDATE;TZID=Europe/Vienna:20130101T000000,20130102T000000" in ical
    assert b"EXDATE;TZID=Europe/Vienna:20130103T000000" in ical



"""Test the sequence and other int properies.

https://www.rfc-editor.org/rfc/rfc5545#section-3.8.7.4
https://github.com/collective/icalendar/issues/802

"""

import pytest

from icalendar.cal.component import Component
from icalendar.cal.event import Event
from icalendar.cal.journal import Journal
from icalendar.cal.todo import Todo


@pytest.fixture(params=[0, None])
def default_sequence(request):
    return request.param


@pytest.fixture(params=[Event, Journal, Todo])
def component(request, default_sequence) -> Component:
    """Return a component."""
    component: Component = request.param()
    if default_sequence is not None:
        component["SEQUENCE"] = default_sequence
    return component


def test_sequence_is_0(component: Component):
    """Check the default value."""
    assert component.sequence == 0


def test_increase_sequence(component: Component):
    """Check the default value."""
    component.sequence += 1
    assert component.sequence == 1
    assert component["SEQUENCE"] == 1


def test_set_sequence(component: Component):
    """Check the default value."""
    component.sequence = 400
    assert component.sequence == 400
    assert component["SEQUENCE"] == 400


def test_delete_sequence_default(component: Component):
    """Delete the value."""
    del component.sequence
    assert component.sequence == 0


def test_delete_sequence_with_value(component: Component):
    """Delete the value."""
    component.sequence = 400
    del component.sequence
    assert component.sequence == 0



"""The RALATED-TO property is extended by :rfc:`9253`.

These are the tests for related_to.
"""

import pytest

from icalendar import (
    RELTYPE,
    Alarm,
    Availability,
    Available,
    Calendar,
    Event,
    FreeBusy,
    Journal,
    Timezone,
    TimezoneDaylight,
    TimezoneStandard,
    Todo,
    vText,
    vUid,
    vUri,
)


def test_examples_from_rfc_9253(calendars):
    """Section 9.1 provides some examples."""
    calendar: Calendar = calendars.rfc_9253_related_to
    event = calendar.events[0]
    related_to = event.related_to

    assert len(related_to) == 3

    assert (
        related_to[0].ical_value == "jsmith.part7.19960817T083000.xyzMail@example.com"
    )
    assert related_to[0].RELTYPE == "PARENT", "default"
    assert isinstance(related_to[0], vText)

    assert related_to[1].uid == "19960401-080045-4000F192713-0052@example.com"
    assert related_to[1].ical_value == "19960401-080045-4000F192713-0052@example.com"
    assert related_to[1].RELTYPE == "PARENT", "default"
    assert isinstance(related_to[1], vUid)

    assert (
        related_to[2].uri
        == "https://example.com/caldav/user/jb/cal/19960401-080045-4000F192713.ics"
    )
    assert (
        related_to[2].ical_value
        == "https://example.com/caldav/user/jb/cal/19960401-080045-4000F192713.ics"
    )
    assert related_to[2].RELTYPE == RELTYPE.STARTTOFINISH
    assert isinstance(related_to[2], vUri)


@pytest.fixture(
    params=(
        Alarm,
        Availability,
        Available,
        Calendar,
        Event,
        FreeBusy,
        Journal,
        Timezone,
        TimezoneDaylight,
        TimezoneStandard,
        Todo,
    )
)
def related_to_component_class(request):
    """The classes to test.

    rfc:`9253`::

        Conformance: RELATED-TO: This property can be specified zero or
        more times in any iCalendar component.

    """
    return request.param


@pytest.fixture
def related_to_component(related_to_component_class):
    """An instance of the component class to test."""
    return related_to_component_class()


def test_no_related_to_by_default(related_to_component):
    """By default, components have no RELATED-TO properties."""
    assert "RELATED-TO" not in related_to_component
    assert related_to_component.related_to == []


def test_strings_are_converted_to_the_default_value_type(related_to_component):
    """We set the URI.

    According to :rfc:`5545`, this should be vText.
    """
    related_to_component.related_to = ["https://example.com"]
    assert related_to_component.related_to == [vText("https://example.com")]
    assert isinstance(related_to_component.related_to[0], vText)
    assert related_to_component.related_to[0].params.value is None, (
        "We do not need to specify the default value."
    )



"""Test common properties of components."""

from datetime import date, datetime, timedelta

import pytest

from icalendar import vDDDTypes
from icalendar.cal.component import Component
from icalendar.cal.event import Event
from icalendar.cal.free_busy import FreeBusy
from icalendar.cal.journal import Journal
from icalendar.cal.todo import Todo
from icalendar.error import InvalidCalendar


@pytest.fixture(params=[Event, Todo, Journal, FreeBusy])
def dtstamp_comp(request):
    """a component to test"""
    return request.param()


def test_no_dtstamp(dtstamp_comp):
    """We have None as a value."""
    assert dtstamp_comp.DTSTAMP is None


def set_dtstamp_attribute(component: Component, value: date):
    """Use the setter."""
    component.DTSTAMP = value


def set_dtstamp_item(component: Component, value: date):
    """Use setitem."""
    component["DTSTAMP"] = vDDDTypes(value)


def set_dtstamp_add(component: Component, value: date):
    """Use add."""
    component.add("DTSTAMP", value)


@pytest.mark.parametrize(
    ("value", "timezone", "expected"),
    [
        (datetime(2024, 10, 11, 23, 1), None, datetime(2024, 10, 11, 23, 1)),
        (datetime(2024, 10, 11, 23, 1), "Europe/Berlin", datetime(2024, 10, 11, 21, 1)),
        (datetime(2024, 10, 11, 22, 1), "UTC", datetime(2024, 10, 11, 22, 1)),
        (date(2024, 10, 10), None, datetime(2024, 10, 10)),
    ],
)
@pytest.mark.parametrize(
    "set_dtstamp", [set_dtstamp_add, set_dtstamp_attribute, set_dtstamp_item]
)
def test_set_value_and_get_it(
    dtstamp_comp, value, timezone, expected, tzp, set_dtstamp
):
    """Set and get the DTSTAMP value."""
    dtstamp = value if timezone is None else tzp.localize(value, timezone)
    set_dtstamp(dtstamp_comp, dtstamp)
    in_utc = tzp.localize_utc(expected)
    get_value = dtstamp_comp.get("DTSTAMP").dt
    assert in_utc == get_value or set_dtstamp != set_dtstamp_attribute
    assert in_utc == dtstamp_comp.DTSTAMP


@pytest.mark.parametrize("invalid_value", ["asdasd", timedelta()])
def test_set_invalid_value(invalid_value, dtstamp_comp):
    """Check handling of invalid values."""
    with pytest.raises(TypeError) as e:
        dtstamp_comp.DTSTAMP = invalid_value
    assert e.value.args[0] == f"DTSTAMP takes a datetime in UTC, not {invalid_value}"


@pytest.mark.parametrize("invalid_value", ["ashdkjasjkdkd", vDDDTypes(timedelta())])
def test_get_invalid_value(invalid_value, dtstamp_comp):
    """Check handling of invalid values."""
    dtstamp_comp["DTSTAMP"] = invalid_value
    with pytest.raises(InvalidCalendar) as e:
        dtstamp_comp.DTSTAMP  # noqa: B018
    assert (
        e.value.args[0]
        == f"DTSTAMP must be a datetime in UTC, not {getattr(invalid_value, 'dt', invalid_value)}"
    )


def test_set_twice(dtstamp_comp, tzp):
    """Set the value twice."""
    dtstamp_comp.DTSTAMP = date(2014, 1, 1)
    dtstamp_comp.DTSTAMP = date(2014, 1, 2)
    assert tzp.localize_utc(datetime(2014, 1, 2)) == dtstamp_comp.DTSTAMP


def test_last_modified(dtstamp_comp, tzp):
    """Check we can set LAST_MODIFIED in the same way as DTSTAMP"""
    dtstamp_comp.LAST_MODIFIED = date(2014, 1, 2)
    assert tzp.localize_utc(datetime(2014, 1, 2)) == dtstamp_comp.LAST_MODIFIED


@pytest.fixture(params=["last_modified", "created"])
def attr_depending_on_dtstamp(request):
    """These parameters default to the DTSTAMP attribute."""
    return request.param


@pytest.mark.parametrize(
    "dt", [datetime(2011, 10, 5, 13, 32, 25), datetime(2011, 10, 5)]
)
def test_last_modified_defaults_to_dtstamp(dt, tzp, c, attr_depending_on_dtstamp):
    """In the case of an iCalendar object that doesn't specify a "METHOD"
    property, "DTSTAMP" property is equivalent to the "LAST-MODIFIED"
    property."""
    c.stamp = dt
    value = getattr(c, attr_depending_on_dtstamp)
    assert value == c.stamp


def test_set_last_modified(c, attr_depending_on_dtstamp):
    """Setting last modified does not override the dtstamp."""
    setattr(c, attr_depending_on_dtstamp, datetime(2011, 10, 5, 13, 32, 25))
    assert c.DTSTAMP is None


def test_set_last_modified_2(c, tzp, attr_depending_on_dtstamp):
    """Setting last modified does not override the dtstamp."""
    c.DTSTAMP = datetime(2011, 10, 5)
    setattr(c, attr_depending_on_dtstamp, datetime(2011, 10, 5, 13, 32, 25))
    assert c.DTSTAMP == tzp.localize_utc(datetime(2011, 10, 5))


def test_delete_last_modified_1(c, attr_depending_on_dtstamp):
    """Delete the absent value."""
    delattr(c, attr_depending_on_dtstamp)
    assert getattr(c, attr_depending_on_dtstamp) is None


def test_delete_last_modified_2(c, attr_depending_on_dtstamp):
    """Delete the present value."""
    setattr(c, attr_depending_on_dtstamp, datetime(2011, 10, 5, 13, 32, 25))
    delattr(c, attr_depending_on_dtstamp)
    assert getattr(c, attr_depending_on_dtstamp) is None



""":rfc:`5545` VTODO component."""

from __future__ import annotations

import uuid
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, Literal, Sequence

from icalendar.attr import (
    CONCEPTS_TYPE_SETTER,
    LINKS_TYPE_SETTER,
    RELATED_TO_TYPE_SETTER,
    X_MOZ_LASTACK_property,
    X_MOZ_SNOOZE_TIME_property,
    attendees_property,
    categories_property,
    class_property,
    color_property,
    conferences_property,
    contacts_property,
    create_single_property,
    description_property,
    exdates_property,
    get_duration_property,
    get_end_property,
    get_start_end_duration_with_validation,
    get_start_property,
    images_property,
    location_property,
    organizer_property,
    priority_property,
    property_del_duration,
    property_doc_duration_template,
    property_get_duration,
    property_set_duration,
    rdates_property,
    rrules_property,
    sequence_property,
    set_duration_with_locking,
    set_end_with_locking,
    set_start_with_locking,
    status_property,
    summary_property,
    uid_property,
    url_property,
)
from icalendar.cal.component import Component

if TYPE_CHECKING:
    from icalendar.alarms import Alarms
    from icalendar.enums import CLASS, STATUS
    from icalendar.prop import vCalAddress
    from icalendar.prop.conference import Conference


class Todo(Component):
    """
    A "VTODO" calendar component is a grouping of component
    properties that represents an action item or assignment. For
    example, it can be used to represent an item of work assigned to
    an individual, such as "Prepare for the upcoming conference
    seminar on Internet Calendaring".

    Examples:
        Create a new Todo:

            >>> from icalendar import Todo
            >>> todo = Todo.new()
            >>> print(todo.to_ical())
            BEGIN:VTODO
            DTSTAMP:20250517T080612Z
            UID:d755cef5-2311-46ed-a0e1-6733c9e15c63
            END:VTODO

    """

    name = "VTODO"

    required = (
        "UID",
        "DTSTAMP",
    )
    singletons = (
        "CLASS",
        "COLOR",
        "COMPLETED",
        "CREATED",
        "DESCRIPTION",
        "DTSTAMP",
        "DTSTART",
        "GEO",
        "LAST-MODIFIED",
        "LOCATION",
        "ORGANIZER",
        "PERCENT-COMPLETE",
        "PRIORITY",
        "RECURRENCE-ID",
        "SEQUENCE",
        "STATUS",
        "SUMMARY",
        "UID",
        "URL",
        "DUE",
        "DURATION",
    )
    exclusive = (
        "DUE",
        "DURATION",
    )
    multiple = (
        "ATTACH",
        "ATTENDEE",
        "CATEGORIES",
        "COMMENT",
        "CONTACT",
        "EXDATE",
        "RSTATUS",
        "RELATED",
        "RESOURCES",
        "RDATE",
        "RRULE",
    )
    DTSTART = create_single_property(
        "DTSTART",
        "dt",
        (datetime, date),
        date,
        'The "DTSTART" property for a "VTODO" specifies the inclusive start of the Todo.',  # noqa: E501
    )
    DUE = create_single_property(
        "DUE",
        "dt",
        (datetime, date),
        date,
        'The "DUE" property for a "VTODO" calendar component specifies the non-inclusive end of the Todo.',  # noqa: E501
    )
    DURATION = property(
        property_get_duration,
        property_set_duration,
        property_del_duration,
        property_doc_duration_template.format(component="VTODO"),
    )

    def _get_start_end_duration(self):
        """Verify the calendar validity and return the right attributes."""
        return get_start_end_duration_with_validation(self, "DTSTART", "DUE", "VTODO")

    @property
    def start(self) -> date | datetime:
        """The start of the VTODO.

        Invalid values raise an InvalidCalendar.
        If there is no start, we also raise an IncompleteComponent error.

        You can get the start, end and duration of a Todo as follows:

        >>> from datetime import datetime
        >>> from icalendar import Todo
        >>> todo = Todo()
        >>> todo.start = datetime(2021, 1, 1, 12)
        >>> todo.end = datetime(2021, 1, 1, 12, 30) # 30 minutes
        >>> todo.duration  # 1800 seconds == 30 minutes
        datetime.timedelta(seconds=1800)
        >>> print(todo.to_ical())
        BEGIN:VTODO
        DTSTART:20210101T120000
        DUE:20210101T123000
        END:VTODO
        """
        return get_start_property(self)

    @start.setter
    def start(self, start: date | datetime | None):
        """Set the start."""
        self.DTSTART = start

    @property
    def end(self) -> date | datetime:
        """The end of the todo.

        Invalid values raise an InvalidCalendar error.
        If there is no end, we also raise an IncompleteComponent error.
        """
        return get_end_property(self, "DUE")

    @end.setter
    def end(self, end: date | datetime | None):
        """Set the end."""
        self.DUE = end

    @property
    def duration(self) -> timedelta:
        """The duration of the VTODO.

        Returns the DURATION property if set, otherwise calculated from start and end.
        You can set the duration to automatically adjust the end time while keeping
        start locked.

        Setting the duration will:
        1. Keep the start time locked (unchanged)
        2. Adjust the end time to start + duration
        3. Remove any existing DUE property
        4. Set the DURATION property
        """
        return get_duration_property(self)

    @duration.setter
    def duration(self, value: timedelta):
        if not isinstance(value, timedelta):
            raise TypeError(f"Use timedelta, not {type(value).__name__}.")

        # Use the set_duration method with default start-locked behavior
        self.set_duration(value, locked="start")

    def set_duration(
        self, duration: timedelta | None, locked: Literal["start", "end"] = "start"
    ):
        """Set the duration of the event relative to either start or end.

        Args:
            duration: The duration to set, or None to convert to DURATION property
            locked: Which property to keep unchanged ('start' or 'end')
        """
        set_duration_with_locking(self, duration, locked, "DUE")

    def set_start(
        self, start: date | datetime, locked: Literal["duration", "end"] | None = None
    ):
        """Set the start with explicit locking behavior.

        Args:
            start: The start time to set
            locked: Which property to keep unchanged ('duration', 'end', or None
                for auto-detect)
        """
        set_start_with_locking(self, start, locked, "DUE")

    def set_end(
        self, end: date | datetime, locked: Literal["start", "duration"] = "start"
    ):
        """Set the end of the component, keeping either the start or the duration same.

        Args:
            end: The end time to set
            locked: Which property to keep unchanged ('start' or 'duration')
        """
        set_end_with_locking(self, end, locked, "DUE")

    X_MOZ_SNOOZE_TIME = X_MOZ_SNOOZE_TIME_property
    X_MOZ_LASTACK = X_MOZ_LASTACK_property

    @property
    def alarms(self) -> Alarms:
        """Compute the alarm times for this component.

        >>> from datetime import datetime
        >>> from icalendar import Todo
        >>> todo = Todo()  # empty without alarms
        >>> todo.start = datetime(2024, 10, 26, 10, 21)
        >>> len(todo.alarms.times)
        0

        Note that this only uses DTSTART and DUE, but ignores
        RDATE, EXDATE, and RRULE properties.
        """
        from icalendar.alarms import Alarms

        return Alarms(self)

    color = color_property
    sequence = sequence_property
    categories = categories_property
    rdates = rdates_property
    exdates = exdates_property
    rrules = rrules_property
    uid = uid_property
    summary = summary_property
    description = description_property
    classification = class_property
    url = url_property
    organizer = organizer_property
    location = location_property
    priority = priority_property
    contacts = contacts_property
    status = status_property
    attendees = attendees_property
    images = images_property
    conferences = conferences_property

    @classmethod
    def new(
        cls,
        /,
        attendees: list[vCalAddress] | None = None,
        categories: Sequence[str] = (),
        classification: CLASS | None = None,
        color: str | None = None,
        comments: list[str] | str | None = None,
        concepts: CONCEPTS_TYPE_SETTER = None,
        contacts: list[str] | str | None = None,
        conferences: list[Conference] | None = None,
        created: date | None = None,
        description: str | None = None,
        end: date | datetime | None = None,
        last_modified: date | None = None,
        links: LINKS_TYPE_SETTER = None,
        location: str | None = None,
        organizer: vCalAddress | str | None = None,
        priority: int | None = None,
        refids: list[str] | str | None = None,
        related_to: RELATED_TO_TYPE_SETTER = None,
        sequence: int | None = None,
        stamp: date | None = None,
        start: date | datetime | None = None,
        status: STATUS | None = None,
        summary: str | None = None,
        uid: str | uuid.UUID | None = None,
        url: str | None = None,
    ):
        """Create a new TODO with all required properties.

        This creates a new Todo in accordance with :rfc:`5545`.

        Arguments:
            attendees: The :attr:`attendees` of the todo.
            categories: The :attr:`categories` of the todo.
            classification: The :attr:`classification` of the todo.
            color: The :attr:`color` of the todo.
            comments: The :attr:`~icalendar.Component.comments` of the todo.
            concepts: The :attr:`~icalendar.Component.concepts` of the todo.
            contacts: The :attr:`contacts` of the todo.
            conferences: The :attr:`conferences` of the todo.
            created: The :attr:`~icalendar.Component.created` of the todo.
            description: The :attr:`description` of the todo.
            end: The :attr:`end` of the todo.
            last_modified: The :attr:`~icalendar.Component.last_modified` of the todo.
            links: The :attr:`~icalendar.Component.links` of the todo.
            location: The :attr:`location` of the todo.
            organizer: The :attr:`organizer` of the todo.
            refids: :attr:`~icalendar.Component.refids` of the todo.
            related_to: :attr:`~icalendar.Component.related_to` of the todo.
            sequence: The :attr:`sequence` of the todo.
            stamp: The :attr:`~icalendar.Component.DTSTAMP` of the todo.
                If None, this is set to the current time.
            start: The :attr:`start` of the todo.
            status: The :attr:`status` of the todo.
            summary: The :attr:`summary` of the todo.
            uid: The :attr:`uid` of the todo.
                If None, this is set to a new :func:`uuid.uuid4`.
            url: The :attr:`url` of the todo.

        Returns:
            :class:`Todo`

        Raises:
            InvalidCalendar: If the content is not valid according to :rfc:`5545`.

        .. warning:: As time progresses, we will be stricter with the validation.
        """
        todo: Todo = super().new(
            stamp=stamp if stamp is not None else cls._utc_now(),
            created=created,
            last_modified=last_modified,
            comments=comments,
            links=links,
            related_to=related_to,
            refids=refids,
            concepts=concepts,
        )
        todo.summary = summary
        todo.description = description
        todo.uid = uid if uid is not None else uuid.uuid4()
        todo.start = start
        todo.end = end
        todo.color = color
        todo.categories = categories
        todo.sequence = sequence
        todo.classification = classification
        todo.url = url
        todo.organizer = organizer
        todo.location = location
        todo.priority = priority
        todo.contacts = contacts
        todo.status = status
        todo.attendees = attendees
        todo.conferences = conferences

        if cls._validate_new:
            cls._validate_start_and_end(start, end)
        return todo


__all__ = ["Todo"]



"""Check some module consistency."""

from datetime import date, datetime, time, timedelta

import pytest

import icalendar
from icalendar import Parameters, prop, vDDDLists, vDDDTypes
from icalendar.tests.data import PARAMETER_NAMES


def test_all_propeties_are_exported(v_prop_name):
    """icalendar.prop should export all these properties."""
    assert v_prop_name in prop.__all__
    assert v_prop_name in icalendar.__all__
    assert getattr(prop, v_prop_name, None) is not None
    assert getattr(icalendar, v_prop_name, None) is not None


def test_all_properties_have_an_examples_method(v_prop):
    """All properties have a class method that returns a list of examples."""
    examples = v_prop.examples()
    assert isinstance(examples, list), "We expect a list of examples."
    assert len(examples) > 0, "We have examples."
    assert all(isinstance(example, v_prop) for example in examples), (
        "All examples should be of that type."
    )
    assert v_prop.__name__ in v_prop.examples.__doc__.splitlines()[0], (
        "type is in docstring"
    )


def test_all_properties_are_part_of_the_union_type(v_prop):
    """Check the union type captures all vProperty instances."""
    assert v_prop in prop.VPROPERTY.__args__


@pytest.mark.parametrize("no_prop", PARAMETER_NAMES)
def test_if_not_part_of_properties_they_are_not_included_in_the_type(no_prop):
    """If these are not properties rather parameters, do not include them."""
    assert no_prop not in prop.VPROPERTY.__args__


default_value_map = {
    prop.vBoolean: "BOOLEAN",
    prop.vCalAddress: "CAL-ADDRESS",
    prop.vCategory: "TEXT",
    prop.vDate: "DATE",
    prop.vDatetime: "DATE-TIME",
    prop.vDuration: "DURATION",
    prop.vFloat: "FLOAT",
    prop.vInt: "INTEGER",
    prop.vPeriod: "PERIOD",
    prop.vRecur: "RECUR",
    prop.vText: "TEXT",
    prop.vTime: "TIME",
    prop.vUTCOffset: "UTC-OFFSET",
    prop.vUri: "URI",
    prop.vBinary: "BINARY",
    prop.vGeo: "FLOAT",
}


@pytest.mark.parametrize(("v_prop", "default_value"), list(default_value_map.items()))
def test_get_default_value(v_prop_example: prop.VPROPERTY, default_value: str):
    """Check the default value of all properties."""
    del v_prop_example.VALUE
    assert v_prop_example.VALUE == default_value
    assert "VALUE" not in v_prop_example.params


def test_set_value(v_prop_example):
    """Test setting the VALUE parameter."""
    v_prop_example.VALUE = "X-OTHER-VALUE"
    assert v_prop_example.VALUE == "X-OTHER-VALUE"
    assert v_prop_example.params.value == "X-OTHER-VALUE"


def test_delete_the_set_value(v_prop_example, v_prop):
    """The value is defaulting after delete."""
    v_prop_example.VALUE = "X-VALUE"
    del v_prop_example.VALUE
    assert v_prop_example.VALUE != "X-VALUE"
    if v_prop in default_value_map:
        assert v_prop_example.VALUE == default_value_map[v_prop], (
            "deleted value defaults"
        )


@pytest.mark.parametrize("v_prop", [vDDDLists, vDDDTypes])
@pytest.mark.parametrize(
    ("dt", "value"),
    [
        (datetime(2018, 1, 1), "DATE-TIME"),
        (date(2018, 1, 1), "DATE"),
        (time(1, 1), "TIME"),
        (timedelta(1, 1), "DURATION"),
    ],
)
def test_dt_value(v_prop, dt, value):
    """Check that the VALUE parameter is correctly determined."""
    assert v_prop(dt).VALUE == value


def test_special_case_no_dts():
    """If the list is empty, we still need a value."""
    assert vDDDLists([]).VALUE == "DATE-TIME"


def special_case_period():
    """Check PERIOD for vDDDLists and vDDDTypes."""
    assert vDDDLists([(datetime(2018, 1, 1), timedelta(1, 1))]).VALUE == "PERIOD"
    assert vDDDTypes((datetime(2018, 1, 1), timedelta(1, 1))).VALUE == "PERIOD"


def test_value_parameter_does_not_turn_up_in_jcal():
    """The VALUE parameter should not turn up if set."""
    params = Parameters()
    params["VALUE"] = "DATE-TIME"
    assert params.to_jcal() == {}


def test_value_is_always_uppercase(v_prop_example):
    """The VALUE parameter should always be uppercase."""
    v_prop_example.VALUE = "unknown"
    assert v_prop_example.VALUE == "UNKNOWN"


def test_setting_the_value_turns_it_uppercase():
    """The VALUE parameter should always be uppercase."""
    parameters = Parameters()
    parameters.value = "unknown"
    assert parameters.value == "UNKNOWN"


def test_default_value_is_always_uppercase():
    """The default value should always be uppercase."""
    pytest.skip("TODO at a later stage when the __init__ function is implemented.")
    parameters = Parameters({"VALUE": "unknown1"})
    assert parameters.value == "UNKNOWN1"
    assert parameters["VALUE"] == "UNKNOWN1"


@pytest.mark.parametrize("attribute", ["VALUE", "to_jcal", "to_ical"])
def test_common_methods(v_prop, attribute):
    """Check that common functionality is provided by all of them."""
    assert hasattr(v_prop, attribute), (
        f'{v_prop.__name__} is missing a function or property named "{attribute}".'
    )


def test_all_possible_value_types_are_listed(types_factory):
    """Check that all values types added are also part of the TypesFactory."""
    for value in types_factory.values():
        assert value in prop.VPROPERTY.__args__, (
            f"The value type {value} is not part of VPROPERTY."
        )
        assert value in types_factory.all_types, (
            f"The value type {value} is not part of TypesFactory."
        )


def test_all_value_types_are_in_the_factory(v_prop, types_factory):
    """Check that all vProperty types are part of the TypesFactory."""
    assert v_prop in types_factory.all_types, (
        f"The value type {v_prop} is not part of TypesFactory.all_types."
    )


def test_an_new_component_class_is_registered(component_factory):
    """Check that the result is always the same."""
    assert component_factory.get_component_class(
        "VCUSTOM"
    ) is component_factory.get_component_class("VCUSTOM")



"""A factory to create components."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from icalendar.caselessdict import CaselessDict

if TYPE_CHECKING:
    from icalendar import Component


class ComponentFactory(CaselessDict):
    """Registered components from :rfc:`7953` and :rfc:`5545`.

    To get a component, use this class as shown below.

    .. code-block:: pycon

        >>> from icalendar import ComponentFactory
        >>> factory = ComponentFactory()
        >>> event_class = factory.get_component_class('VEVENT')
        >>> event_class()
        VEVENT({})

    If a component class is not yet supported, it can be either created
    using :meth:`get_component_class` or added manually as a subclass of
    :class:`~icalendar.Component`.
    """

    def __init__(self, *args, **kwargs):
        """Set keys to upper for initial dict."""
        super().__init__(*args, **kwargs)
        from icalendar.cal.alarm import Alarm
        from icalendar.cal.availability import Availability
        from icalendar.cal.available import Available
        from icalendar.cal.calendar import Calendar
        from icalendar.cal.event import Event
        from icalendar.cal.free_busy import FreeBusy
        from icalendar.cal.journal import Journal
        from icalendar.cal.timezone import Timezone, TimezoneDaylight, TimezoneStandard
        from icalendar.cal.todo import Todo

        self.add_component_class(Calendar)
        self.add_component_class(Event)
        self.add_component_class(Todo)
        self.add_component_class(Journal)
        self.add_component_class(FreeBusy)
        self.add_component_class(Timezone)
        self.add_component_class(TimezoneStandard)
        self.add_component_class(TimezoneDaylight)
        self.add_component_class(Alarm)
        self.add_component_class(Available)
        self.add_component_class(Availability)

    def add_component_class(self, cls: type[Component]) -> None:
        """Add a component class to the factory.

        Args:
            cls: The component class to add.
        """
        self[cls.name] = cls

    def get_component_class(self, name: str) -> type[Component]:
        """Get the component class from the factory.

        This also creates and adds the component class if it does not exist.

        Args:
            name: The name of the component, for example, ``"VCALENDAR"``.

        Returns:
            The registered component class.
        """
        component_class = self.get(name)
        if component_class is None:
            from icalendar.cal.component import Component

            component_class = type(
                re.sub(r"[^\w]+", "", name), (Component,), {"name": name.upper()}
            )
            self.add_component_class(component_class)
        return component_class


__all__ = ["ComponentFactory"]



import pytest

from icalendar.prop import vWeekday


def test_simple():
    weekday = vWeekday("SU")
    assert weekday.to_ical() == b"SU"
    assert weekday.weekday == "SU"
    assert weekday.relative is None


def test_relative():
    weekday = vWeekday("-1MO")
    assert weekday.to_ical() == b"-1MO"
    assert weekday.weekday == "MO"
    assert weekday.relative == -1


def test_roundtrip():
    assert vWeekday.from_ical(vWeekday("+2TH").to_ical()) == "+2TH"


def test_error():
    """Error: Expected weekday abbrevation, got: \"-100MO\" """
    with pytest.raises(ValueError):
        vWeekday.from_ical("-100MO")



"""Quoting the TZID parameter creates compatibility problems.

See https://github.com/collective/icalendar/issues/836

:rfc:`5545`:

.. code-block:: text

    param-value   = paramtext / quoted-string
    paramtext     = *SAFE-CHAR
    quoted-string = DQUOTE *QSAFE-CHAR DQUOTE
    SAFE-CHAR     = WSP / %x21 / %x23-2B / %x2D-39 / %x3C-7E
                   / NON-US-ASCII
     ; Any character except CONTROL, DQUOTE, ";", ":", ","
     NON-US-ASCII  = UTF8-2 / UTF8-3 / UTF8-4
     ; UTF8-2, UTF8-3, and UTF8-4 are defined in [RFC3629]

"""
from datetime import datetime, time
from icalendar import Event, Parameters, vDDDTypes, vDatetime, vPeriod, vTime
import pytest

from icalendar.prop import vDDDLists

# All the controls except HTAB
CONTROL = {
    i for i in range(256) if 0x00 <= i <= 0x08 or 0x0A <= i <= 0x1F or i == 0x7F
}

# Any character except CONTROL, DQUOTE, ";", ":", ","
SAFE_CHAR = set(range(256)) - CONTROL - set(b'";:,')

param_tzid = pytest.mark.parametrize(
    "tzid",
    [
        "Europe/London",
        "Eastern Standard Time",
    ]
)

@param_tzid
@pytest.mark.parametrize(
    "vdt",
    [
        vDatetime(datetime(2024, 10, 11, 12, 0)),
        vTime(time(23, 59, 59)),
        vDDDTypes(datetime(2024, 10, 11, 12, 0)),
        vPeriod((datetime(2024, 10, 11, 12, 0), datetime(2024, 10, 11, 13, 0))),
        vDDDLists([datetime(2024, 10, 11, 12, 0)]),
    ]
)
def test_parameter_is_not_quoted_when_not_needed(tzid, vdt):
    """Check that serializing the value works without quoting."""
    e = Event()
    vdt.params["TZID"] = tzid
    e["DTSTART"] = vdt
    ics = e.to_ical().decode()
    print(ics)
    assert tzid in ics
    assert f'"{tzid}' not in ics
    assert f'{tzid}"' not in ics


@pytest.mark.parametrize(
    "safe_char",
    list(map(chr, sorted(SAFE_CHAR)))
)
def test_safe_char_is_not_escaped(safe_char):
    """Check that paramerter serialization is without quotes for safe chars."""
    params = Parameters(tzid=safe_char)
    result = params.to_ical().decode()
    assert '"'  not in result


def test_get_calendar_and_serialize_it_wihtout_quotes(calendars):
    """The example calendar should not contain the timezone with quotes."""
    ics = calendars.issue_836_do_not_quote_tzid.to_ical().decode()
    assert '"Eastern Standard' not in ics
    assert 'Standard Time"' not in ics
    assert "Eastern Standard Time" in ics



"""Utility functions for icalendar."""

from __future__ import annotations

from datetime import date, datetime, tzinfo
from typing import Union, cast

try:
    from typing import TypeIs
except ImportError:
    from typing_extensions import TypeIs


def is_date(dt: Union[date, datetime]) -> bool:
    """Whether this is a date and not a datetime."""
    return isinstance(dt, date) and not isinstance(dt, datetime)


def is_datetime(dt: Union[date, datetime]) -> TypeIs[datetime]:
    """Whether this is a datetime and not just a date."""
    return isinstance(dt, datetime)


def to_datetime(dt: Union[date, datetime]) -> datetime:
    """Make sure we have a datetime, not a date."""
    if is_date(dt):
        return datetime(dt.year, dt.month, dt.day)  # noqa: DTZ001
    return cast("datetime", dt)


def is_pytz(tz: tzinfo) -> bool:
    """Whether the timezone requires localize() and normalize()."""
    return hasattr(tz, "localize")


def is_pytz_dt(dt: Union[date, datetime]) -> TypeIs[datetime]:
    """Whether the time requires localize() and normalize()."""
    return is_datetime(dt) and (tzinfo := dt.tzinfo) is not None and is_pytz(tzinfo)


def normalize_pytz(dt: Union[date, datetime]) -> Union[date, datetime]:
    """We have to normalize the time after a calculation if we use pytz.

    pytz requires this function to be used in order to correctly calculate the
    timezone's offset after calculations.
    """
    if is_pytz_dt(dt):
        return dt.tzinfo.normalize(dt)  # type: ignore[attr-defined]
    return dt


__all__ = [
    "is_date",
    "is_datetime",
    "is_pytz",
    "is_pytz_dt",
    "normalize_pytz",
    "to_datetime",
]



""":rfc:`5545` VALARM component."""

from __future__ import annotations

from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, NamedTuple, Optional, Union

from icalendar.attr import (
    CONCEPTS_TYPE_SETTER,
    LINKS_TYPE_SETTER,
    RELATED_TO_TYPE_SETTER,
    attendees_property,
    create_single_property,
    description_property,
    property_del_duration,
    property_get_duration,
    property_set_duration,
    single_int_property,
    single_string_property,
    single_utc_property,
    summary_property,
    uid_property,
)
from icalendar.cal.component import Component

if TYPE_CHECKING:
    import uuid

    from icalendar.prop import vCalAddress


class Alarm(Component):
    """
    A "VALARM" calendar component is a grouping of component
    properties that defines an alarm or reminder for an event or a
    to-do. For example, it may be used to define a reminder for a
    pending event or an overdue to-do.
    """

    name = "VALARM"
    # some properties MAY/MUST/MUST NOT appear depending on ACTION value
    required = (
        "ACTION",
        "TRIGGER",
    )
    singletons = (
        "ATTACH",
        "ACTION",
        "DESCRIPTION",
        "SUMMARY",
        "TRIGGER",
        "DURATION",
        "REPEAT",
        "UID",
        "PROXIMITY",
        "ACKNOWLEDGED",
    )
    inclusive = (
        (
            "DURATION",
            "REPEAT",
        ),
        (
            "SUMMARY",
            "ATTENDEE",
        ),
    )
    multiple = ("ATTENDEE", "ATTACH", "RELATED-TO")

    REPEAT = single_int_property(
        "REPEAT",
        0,
        """The REPEAT property of an alarm component.

        The alarm can be defined such that it triggers repeatedly.  A
        definition of an alarm with a repeating trigger MUST include both
        the "DURATION" and "REPEAT" properties.  The "DURATION" property
        specifies the delay period, after which the alarm will repeat.
        The "REPEAT" property specifies the number of additional
        repetitions that the alarm will be triggered.  This repetition
        count is in addition to the initial triggering of the alarm.
        """,
    )

    DURATION = property(
        property_get_duration,
        property_set_duration,
        property_del_duration,
        """The DURATION property of an alarm component.

    The alarm can be defined such that it triggers repeatedly.  A
    definition of an alarm with a repeating trigger MUST include both
    the "DURATION" and "REPEAT" properties.  The "DURATION" property
    specifies the delay period, after which the alarm will repeat.
    """,
    )

    ACKNOWLEDGED = single_utc_property(
        "ACKNOWLEDGED",
        """This is defined in RFC 9074:

    Purpose: This property specifies the UTC date and time at which the
    corresponding alarm was last sent or acknowledged.

    This property is used to specify when an alarm was last sent or acknowledged.
    This allows clients to determine when a pending alarm has been acknowledged
    by a calendar user so that any alerts can be dismissed across multiple devices.
    It also allows clients to track repeating alarms or alarms on recurring events or
    to-dos to ensure that the right number of missed alarms can be tracked.

    Clients SHOULD set this property to the current date-time value in UTC
    when a calendar user acknowledges a pending alarm. Certain kinds of alarms,
    such as email-based alerts, might not provide feedback as to when the calendar user
    sees them. For those kinds of alarms, the client SHOULD set this property
    when the alarm is triggered and the action is successfully carried out.

    When an alarm is triggered on a client, clients can check to see if an "ACKNOWLEDGED"
    property is present. If it is, and the value of that property is greater than or
    equal to the computed trigger time for the alarm, then the client SHOULD NOT trigger
    the alarm. Similarly, if an alarm has been triggered and
    an "alert" has been presented to a calendar user, clients can monitor
    the iCalendar data to determine whether an "ACKNOWLEDGED" property is added or
    changed in the alarm component. If the value of any "ACKNOWLEDGED" property
    in the alarm changes and is greater than or equal to the trigger time of the alarm,
    then clients SHOULD dismiss or cancel any "alert" presented to the calendar user.
    """,  # noqa: E501
    )

    TRIGGER = create_single_property(
        "TRIGGER",
        "dt",
        (datetime, timedelta),
        Optional[Union[timedelta, datetime]],
        """Purpose:  This property specifies when an alarm will trigger.

    Value Type:  The default value type is DURATION.  The value type can
    be set to a DATE-TIME value type, in which case the value MUST
    specify a UTC-formatted DATE-TIME value.

    Either a positive or negative duration may be specified for the
    "TRIGGER" property.  An alarm with a positive duration is
    triggered after the associated start or end of the event or to-do.
    An alarm with a negative duration is triggered before the
    associated start or end of the event or to-do.""",
    )

    @property
    def TRIGGER_RELATED(self) -> str:  # noqa: N802
        """The RELATED parameter of the TRIGGER property.

        Values are either "START" (default) or "END".

        A value of START will set the alarm to trigger off the
        start of the associated event or to-do.  A value of END will set
        the alarm to trigger off the end of the associated event or to-do.

        In this example, we create an alarm that triggers two hours after the
        end of its parent component:

        >>> from icalendar import Alarm
        >>> from datetime import timedelta
        >>> alarm = Alarm()
        >>> alarm.TRIGGER = timedelta(hours=2)
        >>> alarm.TRIGGER_RELATED = "END"
        """
        trigger = self.get("TRIGGER")
        if trigger is None:
            return "START"
        return trigger.params.get("RELATED", "START")

    @TRIGGER_RELATED.setter
    def TRIGGER_RELATED(self, value: str):  # noqa: N802
        """Set "START" or "END"."""
        trigger = self.get("TRIGGER")
        if trigger is None:
            raise ValueError(
                "You must set a TRIGGER before setting the RELATED parameter."
            )
        trigger.params["RELATED"] = value

    class Triggers(NamedTuple):
        """The computed times of alarm triggers.

        start - triggers relative to the start of the Event or Todo (timedelta)

        end - triggers relative to the end of the Event or Todo (timedelta)

        absolute - triggers at a datetime in UTC
        """

        start: tuple[timedelta]
        end: tuple[timedelta]
        absolute: tuple[datetime]

    @property
    def triggers(self):
        """The computed triggers of an Alarm.

        This takes the TRIGGER, DURATION and REPEAT properties into account.

        Here, we create an alarm that triggers 3 times before the start of the
        parent component:

        >>> from icalendar import Alarm
        >>> from datetime import timedelta
        >>> alarm = Alarm()
        >>> alarm.TRIGGER = timedelta(hours=-4)  # trigger 4 hours before START
        >>> alarm.DURATION = timedelta(hours=1)  # after 1 hour trigger again
        >>> alarm.REPEAT = 2  # trigger 2 more times
        >>> alarm.triggers.start == (timedelta(hours=-4),  timedelta(hours=-3),  timedelta(hours=-2))
        True
        >>> alarm.triggers.end
        ()
        >>> alarm.triggers.absolute
        ()
        """  # noqa: E501
        start = []
        end = []
        absolute = []
        trigger = self.TRIGGER
        if trigger is not None:
            if isinstance(trigger, date):
                absolute.append(trigger)
                add = absolute
            elif self.TRIGGER_RELATED == "START":
                start.append(trigger)
                add = start
            else:
                end.append(trigger)
                add = end
            duration = self.DURATION
            if duration is not None:
                for _ in range(self.REPEAT):
                    add.append(add[-1] + duration)
        return self.Triggers(
            start=tuple(start), end=tuple(end), absolute=tuple(absolute)
        )

    uid = single_string_property(
        "UID",
        uid_property.__doc__,
        "X-ALARMUID",
    )
    summary = summary_property
    description = description_property
    attendees = attendees_property

    @classmethod
    def new(
        cls,
        /,
        attendees: Optional[list[vCalAddress]] = None,
        concepts: CONCEPTS_TYPE_SETTER = None,
        description: Optional[str] = None,
        links: LINKS_TYPE_SETTER = None,
        refids: list[str] | str | None = None,
        related_to: RELATED_TO_TYPE_SETTER = None,
        summary: Optional[str] = None,
        uid: Optional[str | uuid.UUID] = None,
    ):
        """Create a new alarm with all required properties.

        This creates a new Alarm in accordance with :rfc:`5545`.

        Arguments:
            attendees: The :attr:`attendees` of the alarm.
            concepts: The :attr:`~icalendar.Component.concepts` of the alarm.
            description: The :attr:`description` of the alarm.
            links: The :attr:`~icalendar.Component.links` of the alarm.
            refids: :attr:`~icalendar.Component.refids` of the alarm.
            related_to: :attr:`~icalendar.Component.related_to` of the alarm.
            summary: The :attr:`summary` of the alarm.
            uid: The :attr:`uid` of the alarm.

        Returns:
            :class:`Alarm`

        Raises:
            InvalidCalendar: If the content is not valid according to :rfc:`5545`.

        .. warning:: As time progresses, we will be stricter with the validation.
        """
        alarm: Alarm = super().new(
            links=links,
            related_to=related_to,
            refids=refids,
            concepts=concepts,
        )
        alarm.summary = summary
        alarm.description = description
        alarm.uid = uid
        alarm.attendees = attendees
        return alarm


__all__ = ["Alarm"]
